#all.data.list=simudata;
#ncluster.method="Tibs2001SEmax";
#dmat.method="manhattan.meanscore";
#wt.eu=wt;
#K.max=K.max;
#arg.mis=F;
#verbose=F


##############################################################################
### Hierarchical Clustering of Regimens Based on Distance in Means
##############################################################################

#################
## input:
## all.data.list: immune response data for multiple reigmen
# ncluster.method = method to determine the # of clusters. Detail:
#   "firstSEmax","Tibs2001SEmax","globalSEmax","firstmax","globalmax": see clusGap() for details
# dmat.method = method to calculate the dissimilarity matrix. Detail:
#   "-log10p":     -log10 of the pairwise global p-values, using all data from all classes
#   "-log10p.cv":  -log10 of the pairwise global p-values, using data from selected classes by cv.glmnet()
#   "euclidean.meanscore":   weighted euclidean distance based on vector of means between any two arms
#   "manhattan.meanscore":   weighted manhattan distance based on vector of means between any two arms
#   "euclidean.meandist":    Euclidean distance between any two pair of individuals one from each arm, the compute a weighted average
# wt.eu = a vector of class specific weights if dmat.method == "euclidean"  (default 1 for every class)
# nmax = the maximum number of clusters to consider
# arg.mis: indicator for existence of missing data
# verbose = logical indicator (default F)


################
## output
## den=den: object of dendagram
## gfun=gfun: object of gap statistics
## dmat=dmat: distance matrix
##  cluster.idx.full = a vector of membership of clusters (without restriction on the maximum # of clusters)
##  cluster.idx.force = a vector of membership of clusters (forcing the maximum # of clusters <= nmax)
## ncluster.full=ncluster.full: estimated number of clusters if no constraints on the maximum possible number of clusters
## ncluster.force=ncluster.force: estimated number of cluster if maximum possible number of clusters is constrained to be nmax

ds.cluster<-function(all.data.list,ncluster.method, dmat.method,wt.eu,nmax,arg.mis,verbose=F){


   # Data measurements
  Narm <- length(all.data.list)             ####### Number of arms
  Nclass <- dim(all.data.list[[1]])[2]      ####### Number of classes

  K.max=Narm

 NN=NULL
  NN.omit=NULL                          ### allow for nonequal observation from each arm
  for (q in 1:length(all.data.list)){
    NN<-c(NN,nrow(all.data.list[[q]]))
    NN.omit=c(NN.omit,nrow(na.omit(all.data.list[[q]])))
  }

  # format the data list into a data frame
  # rdat is a data frame where each column (labelled Class1, Class2, ...) is an outcome and Arm is the study arm
  all.data.matrix <- do.call(rbind,all.data.list)
  if (arg.mis==F) { rdat <- data.frame(Arm=rep(1:Narm,NN.omit),na.omit(all.data.matrix))
  } else rdat <- data.frame(Arm=rep(1:Narm,NN),all.data.matrix)

  colnames(rdat) <- c("Arm", paste("Class",1:Nclass,sep=""))
  # Arm should be a factor
  rdat$Arm <- factor(rdat$Arm, levels=1:Narm, labels=paste("Arm", 1:Narm, sep=""))

  # Class specific weights
  if (is.null(wt.eu)) wt.eu <- rep(1,Nclass)

  # generate dmat
  if (dmat.method == "-log10p") dmat <- myclust.logP(rdat[,-1], Narm, Narm=Narm, Nclass=Nclass, N=N, dmat.out=T)
  if (dmat.method == "euclidean.meanscore") dmat <- myclust.euclidean.meanscore(rdat[,-1], Narm, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, N=N, dmat.out=T)
  if (dmat.method == "manhattan.meanscore") {
    if (arg.mis==F)
    dmat <- myclust.manhattan.meanscore(na.omit(rdat[,-1]), Narm, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN.omit, dmat.out=T)
    else dmat <- myclust.manhattan.meanscore(rdat[,-1], Narm, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=T)
  }
  if (dmat.method == "-log10p.cv") dmat <- myclust.logP.cv(rdat[,-1], Narm, Narm=Narm, Nclass=Nclass, N=N, dmat.out=T)

  # compute the number of clusters
  # ncluster.full = no restriction
  # ncluster.force = restrict the maximum # of clusters to be nmax
  # using -log10 transformed pairwise-test p-values to compute dmat, gap statistic
    if (dmat.method == "-log10p") {
      myclust <- myclust.logP
       if (arg.mis==F) gfun <- try(clusGap(na.omit(rdat[,-1]), FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN.omit, dmat.out=F,verbose=verbose),T)
      else gfun <- try(clusGapNew(rdat[,-1], FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=F,verbose=verbose),T)
    }
    if (dmat.method == "-log10p.cv") {
      myclust <- myclust.logP.cv
        if (arg.mis==F) gfun <- try(clusGap(na.omit(rdat[,-1]), FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN.omit, dmat.out=F,verbose=verbose),T)
      else gfun <- try(clusGapNew(rdat[,-1], FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=F,verbose=verbose),T)
    }
    # using Euclidean distance to compute dmat, gap statistic
    if (dmat.method == "euclidean.meanscore") {
      #set.seed(0)
      myclust <- myclust.euclidean.meanscore
      if (arg.mis==F) gfun <- try(clusGap(na.omit(rdat[,-1]), FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN.omit, dmat.out=F,verbose=verbose),T)
      else gfun <- try(clusGapNew(rdat[,-1], FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=F,verbose=verbose),T)
    }
  if (dmat.method == "manhattan.meanscore") {
      #set.seed(0)
      myclust <- myclust.manhattan.meanscore

      if (arg.mis==F) gfun <- try(clusGap(na.omit(rdat[,-1]), FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN.omit, dmat.out=F,verbose=verbose),T)
      else gfun <- try(clusGapNew(rdat[,-1], FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=F,verbose=verbose),T)
    }

  if (dmat.method == "euclidean.meandist") {
      #set.seed(0)
      myclust <- myclust.euclidean.meandist
      if (arg.mis==F) gfun <- try(clusGap(na.omit(rdat[,-1]), FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN.omit, dmat.out=F,verbose=verbose),T)
      else gfun <- try(clusGapNew(rdat[,-1], FUN=myclust, K.max=K.max, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=F,verbose=verbose),T)
    }



    # determine the # of clusters
    if (class(gfun)=="try-error") {
      list(cluster.idx.force=rep(NA,Narm), cluster.idx.full=rep(NA,Narm),
           ncluster.force=NA, ncluster.full=NA)
     } else {

        ncluster.full <- maxSE(gfun$Tab[,"gap"], gfun$Tab[,"SE.sim"], method=ncluster.method)
        ncluster.force <- maxSE(gfun$Tab[1:nmax,"gap"], gfun$Tab[1:nmax,"SE.sim"], method=ncluster.method)
     }


  # compute dendrogram
  den <- as.dendrogram(hclust(as.dist(dmat)))

  # return a vector with group memberships for restricted clustering
  cluster.idx.force <- cutree(as.hclust(den), k=ncluster.force)
  names(cluster.idx.force) <- paste("Arm",1:Narm,sep="")

  # return a vector with group memberships for non-restrictedd clustering
  cluster.idx.full <- cutree(as.hclust(den), k=ncluster.full)
  names(cluster.idx.full) <- paste("Arm",1:Narm,sep="")

  return(list(den=den,gfun=gfun,dmat=dmat,cluster.idx.full=cluster.idx.full,
  cluster.idx.force=cluster.idx.force,ncluster.full=ncluster.full,ncluster.force=ncluster.force))

}

##########################

###############################################
### Simulation data
###############################################

### Function to generate simulation data

# N = sample size per arm
# Narm = # of arms
# S = covariance matrix (same across arms)
# M = mean vector (each row represents an arm; each column represents a class)

### return a data list of length Narm

### simulate from multivariate normal
simudata.fun <- function(N,Narm,S,M) {
  all.data <- list(NULL)
  for (k in 1:Narm) {
    all.data[[k]] <- mvrnorm(n=N,mu=M[k,],Sigma=S)
  }
  all.data
}




try.error<-function(expr, silent=TRUE) {

  # make it to error
  op <- options("warn")
  on.exit(options(op))
  options(warn = -1)

  # catch using the try() function
  try(expr, silent=TRUE)
}



##### a modified version of clusGap allowing for missing value ##########################
clusGapNew<-function (x, FUNcluster, K.max, B = 100, verbose = interactive(),
    ...)
{
    stopifnot(is.function(FUNcluster), length(dim(x)) == 2, K.max >=
        2, (n <- nrow(x)) >= 1, (p <- ncol(x)) >= 1)
    if (B != (B. <- as.integer(B)) || (B <- B.) <= 0)
        stop("'B' has to be a positive integer")
    if (is.data.frame(x))
        x <- as.matrix(x)
    ii <- seq_len(n)
    W.k <- function(X, kk) {
        clus <- if (kk > 1) {
            FUNcluster(X, kk, ...)$cluster
            #FUNcluster(X, kk, wt.eu=wt.eu, Narm=Narm, Nclass=Nclass, NN=NN, dmat.out=F)$cluster
       } else rep.int(1L, nrow(X))
        0.5 * sum(vapply(split(ii, clus), function(I) {
            xs <- X[I, , drop = FALSE]
            sum(dist(xs)/nrow(xs))
        }, 0))
    }
    logW <- E.logW <- SE.sim <- numeric(K.max)
    if (verbose)
        cat("Clustering k = 1,2,..., K.max (= ", K.max, "): .. ",
            sep = "")
    for (k in 1:K.max) logW[k] <- log(W.k(x, k))
    if (verbose)
        cat("done\n")
    xs <- scale(x, center = TRUE, scale = FALSE)
    m.x <- rep(attr(xs, "scaled:center"), each = n)
    #V.sx <- svd(xs)$v
    V.sx<-svd(na.omit(xs))$v
    rng.x1 <- apply(xs %*% V.sx, 2, range, na.rm=T)
    logWks <- matrix(0, B, K.max)
    if (verbose)
        cat("Bootstrapping, b = 1,2,..., B (= ", B, ")  [one \".\" per sample]:\n",
            sep = "")
    for (b in 1:B) {
        z1 <- apply(rng.x1, 2, function(M, nn) runif(nn, min = M[1],
            max = M[2]), nn = n)
        z <- tcrossprod(z1, V.sx) + m.x
        for (k in 1:K.max) {
            logWks[b, k] <- log(W.k(z, k))
        }
        if (verbose)
            cat(".", if (b%%50 == 0)
                paste(b, "\n"))
    }
    if (verbose && (B%%50 != 0))
        cat("", B, "\n")
    E.logW <- colMeans(logWks)
    SE.sim <- sqrt((1 + 1/B) * apply(logWks, 2, var))
    structure(class = "clusGap", list(Tab = cbind(logW, E.logW,
        gap = E.logW - logW, SE.sim), n = n, B = B, FUNcluster = FUNcluster))
}

###############################################
### Clustering
###############################################

### myclust.euclidean.meanscore(), myclust.euclidean.meandist(), myclust.logP() and myclust.logP.cv()
### are functions called by ds.cluster() to calculate dmat or calculate the gap statistics

### weighted Euclidean distance based on vector of means between any two arms
myclust.euclidean.meanscore <- function(x,k,wt.eu,Narm,Nclass,N,dmat.out=F) {
  mscore <- matrix(NA,Narm,Nclass)
  x <- cbind(rep(1:Narm,each=N),x)
  for ( j in 1:Nclass )
    mscore[,j] <- tapply(x[,j+1],x[,1],mean)
  mscore.wt <- sweep(mscore,2,sqrt(wt.eu),"*")
  dmat <- as.matrix(dist(mscore.wt, method = "euclidean", diag = TRUE,upper=TRUE))
  rownames(dmat) <- levels(x[,1])
  colnames(dmat) <- levels(x[,1])
  arm.cluster=cutree(as.hclust(agnes(dmat, diss=T)), k)
  if (!dmat.out) return(list(cluster=rep(arm.cluster,each=N))) else return(dmat)
}

### weighted mean absolute distance based on vector of means between any two arms
myclust.manhattan.meanscore <- function(x,k,wt.eu,Narm,Nclass,NN,dmat.out=F) {
  mscore <- matrix(NA,Narm,Nclass)
  #x <- cbind(rep(1:Narm,each=N),x)
  x <- cbind(rep(1:Narm,NN),x)

  for ( j in 1:Nclass )
    mscore[,j] <- tapply(x[,j+1],x[,1],mean,na.rm=T)
  mscore.wt <- sweep(mscore,2,sqrt(wt.eu),"*")
  dmat <- as.matrix(dist(mscore.wt, method = "manhattan", diag = TRUE,upper=TRUE))/Nclass
  rownames(dmat) <- levels(x[,1])
  colnames(dmat) <- levels(x[,1])
  arm.cluster=cutree(as.hclust(agnes(dmat, diss=T)), k)
  #if (!dmat.out) return(list(cluster=rep(arm.cluster,each=N))) else return(dmat)
  if (!dmat.out) return(list(cluster=rep(arm.cluster,NN))) else return(dmat)
}
### Euclidean distance between any two pair of individuals one from each arm, the compute a weighted average
myclust.euclidean.meandist <- function(x,k,wt.eu,Narm,Nclass,N,dmat.out=F) {
  ## distance between two arms i,j
  dist2arm.fun <- function(i,j){
    idxi <- data.frame(id=rep(1:N,each=N),order=1:N^2)
    idxj <- data.frame(id=rep(1:N,N),order=1:N^2)
    armi <- data.frame(id=1:N,x2[x2[,1]==i,-1])
    armj <- data.frame(id=1:N,x2[x2[,1]==j,-1])
    scorei <- merge(idxi,armi,by="id",all=T)
    scorej <- merge(idxj,armj,by="id",all=T)
    scorei <- scorei[sort.list(scorei$order),]
    scorej <- scorej[sort.list(scorej$order),]
    d <- sweep((scorei[,-(1:2)] - scorej[,-(1:2)])^2,2,wt.eu,"*")
    mean(sqrt(apply(d,1,sum)))
  }
  x2 <- cbind(rep(1:Narm,each=N),x)
  dmat <- diag(0,Narm,Narm)
  for(i in 1:(Narm-1)){
    for(j in (i+1):Narm){
      dmat[i,j] <- dmat[j,i] <- dist2arm.fun(i,j)
    }
  }
  rownames(dmat) <- levels(x2[,1])
  colnames(dmat) <- levels(x2[,1])
  arm.cluster=cutree(as.hclust(agnes(dmat, diss=T)), k)
  if (!dmat.out) return(list(cluster=rep(arm.cluster,each=N))) else return(dmat)
}

### -log10(p-value)
myclust.logP <- function(x,k,Narm,Nclass,N,dmat.out=F) {
  mscore <- matrix(NA,Narm,Nclass)
  x<-cbind(rep(1:Narm,each=N),x)
  A <- Narm
  pairs <- combn(A, 2)
  pmat <- matrix(1, nr=A, nc=A)
  rownames(pmat) <- levels(x[,1])
  colnames(pmat) <- levels(x[,1])

  for ( i in 1:ncol(pairs) ) {
    a <- unique(x[,1])[pairs[1,i]]
    b <- unique(x[,1])[pairs[2,i]]

    ss <- subset(x, x[,1] %in% c(a,b))

    ss <- as.data.frame(ss)
    colnames(ss)=c('Arm',paste('class',1:Nclass,sep=''))
    uu <- unique(ss[,1])[1]
    fit0 <- glm(I(Arm==uu)~1, data=ss,family=binomial)
    fit1 <- glm(I(Arm==uu)~., data=ss,family=binomial)

    count=0
    while (fit1$null.deviance < fit1$deviance & count<2) {

      ff <- cv.glmnet(x=as.matrix(ss[,-1]),y=I(ss[,1]==uu),family="binomial")
      ind<-(coef(ff)[-1])!=0
      ss=ss[,c(TRUE,as.logical(ind))]
      fit1 <- glm(I(Arm==uu)~., data=ss,family=binomial)
      count=count+1
    }
    pmat[a,b] <- anova(fit0, fit1, test="Chisq")[2,5]
    pmat[b,a] <- pmat[a,b]
    #print(c(i, pmat[a,b]))
  }
  dmat <- -log10(pmat)
  arm.cluster=cutree(as.hclust(agnes(dmat, diss=T)), k)
  if (!dmat.out) return(list(cluster=rep(arm.cluster,each=N))) else return(dmat)
}

### -log10(p-value), with model selection
myclust.logP.cv <- function(x,k,Narm,Nclass,N,dmat.out=F) {
  mscore <- matrix(NA,Narm,Nclass)
  x<-cbind(rep(1:Narm,each=N),x)
  A <- Narm
  pairs <- combn(A, 2)
  pmat <- matrix(1, nr=A, nc=A)
  rownames(pmat) <- levels(x[,1])
  colnames(pmat) <- levels(x[,1])

  for ( i in 1:ncol(pairs) ) {
    a <- unique(x[,1])[pairs[1,i]]
    b <- unique(x[,1])[pairs[2,i]]

    ss <- subset(x, x[,1] %in% c(a,b))
    ss <- as.data.frame(ss)
    colnames(ss)=c('Arm',paste('class',1:Nclass,sep=''))
    uu <- unique(ss[,1])[1]
    fit0 <- glm(I(Arm==uu)~1, data=ss,family=binomial)
    ff <- cv.glmnet(x=as.matrix(ss[,-1]),y=I(ss[,1]==uu),family="binomial")
    ind <- (coef(ff)[-1])!=0
    df <- sum(ind)
    if (df==0) {
      pmat[b,a] <- pmat[a,b] <- 1
    } else{
      ss <- data.frame(ss[,c(TRUE,as.logical(ind))])
      fit1 <- glm(I(Arm==uu)~., data=ss, family="binomial")
      count=0
      while (fit1$null.deviance < fit1$deviance & count<2 & df>0) {
        ff <- cv.glmnet(x=as.matrix(ss[,-1]),y=I(ss[,1]==uu),family="binomial")
        ind <- (coef(ff)[-1])!=0
        df <- sum(ind)
        ss <- as.data.frame(ss[,c(TRUE,as.logical(ind))])
        names(ss) <- c("Arm",as.character(which(ind==T)))
        fit1 <- glm(I(Arm==uu)~., data=ss,family="binomial")
        count=count+1
      }
      if (df == 0) pmat[b,a] <- pmat[a,b] <- 1 else
        pmat[b,a] <- pmat[a,b] <- anova(fit0, fit1, test="Chisq")[2,5]
      #print(c(i, pmat[a,b]))
    }
  }
  dmat <- -log10(pmat)
  arm.cluster=cutree(as.hclust(agnes(dmat, diss=T)), k)
  if (!dmat.out) return(list(cluster=rep(arm.cluster,each=N))) else return(dmat)
}
